/*****************************************************************************
 *
 *  Copyright (C) 2007 Cdric Brgardis <cedric.bregardis AT free.fr> and
 *  Jean-Christian Hassler <jhassler AT free.fr>
 *
 *  This file is part of the Audiowerk2 ALSA driver
 *
 *  The Audiowerk2 ALSA driver is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; version 2.
 *
 *  The Audiowerk2 ALSA driver is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with the Audiowerk2 ALSA driver; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 *****************************************************************************/

#define AW2_SAA7146_M


#include <sound/driver.h>
#include <linux/init.h>
#include <linux/pci.h>
#include <linux/slab.h>
#include <sound/core.h>
#include <sound/initval.h>
#include <sound/pcm.h>
#include <sound/pcm_params.h>

#include <linux/interrupt.h>
#include <asm/io.h>
#include "aw2_tsl.h"
#include "SAA7146.h"

#include <linux/delay.h>



#include "aw2-saa7146.h"


#define WRITEREG(value,addr) writel((value), chip->iobase_virt + (addr)) 
#define READREG(addr) readl(chip->iobase_virt + (addr)) 



static snd_aw2_saa7146_cb_param arrSubstreamItPlaybackCb[NB_STREAM_PLAYBACK];
static snd_aw2_saa7146_cb_param arrSubstreamItCaptureCb[NB_STREAM_CAPTURE];

static int snd_aw2_saa7146_i2c_write(snd_aw2_saa7146_t* chip, unsigned char address, const void* buf, size_t count);
static int snd_aw2_saa7146_get_limit(int size);

void snd_aw2_saa7146_dump_registers(snd_aw2_saa7146_t  *chip)
{
    int i = 0;
    for(i = 0; i <= 0x148; i+=4)
    {
        printk("###DUMP 0x%03x: 0x%08x\n",i,READREG(i));
    }
}


/* chip-specific destructor
 * (see "PCI Resource Managements")
 */
int snd_aw2_saa7146_free(snd_aw2_saa7146_t *chip)
{


    /* disable all irqs */
    WRITEREG(0, IER);


    /* reset saa7146 */
    WRITEREG((MRST_N << 16) , MC1);

    /* release the irq */
    if (chip->irq >= 0)
    {
        free_irq(chip->irq, (void *)chip);
    }
    /* release the i/o ports & memory */
    if (chip->iobase_virt)
    {
        iounmap(chip->iobase_virt);
    }
    pci_release_regions(chip->pci);
    /* disable the PCI entry */
    pci_disable_device(chip->pci);
    /* release the data */
    kfree(chip);
    return 0;
}

void snd_aw2_saa7146_setup(snd_aw2_saa7146_t  *chip)
{
    /* set PCI burst/threshold

    Burst length definition
    VALUE	BURST LENGTH
    000	1 Dword
    001	2 Dwords
    010	4 Dwords
    011	8 Dwords
    100	16 Dwords
    101	32 Dwords
    110	64 Dwords
    111	128 Dwords

    Threshold definition
    VALUE	WRITE MODE		READ MODE
    00	1 Dword of valid data	1 empty Dword
    01	4 Dwords of valid data	4 empty Dwords
    10	8 Dwords of valid data	8 empty Dwords
    11	16 Dwords of valid data	16 empty Dwords */

    unsigned int acon2;
    unsigned int acon1 = 0;
    int i; 


    printk(KERN_ERR "presetup\n");

    /* disable all irqs */
    WRITEREG(0, IER);

    printk(KERN_ERR "reset chip3\n");

    /* reset saa7146 */
    WRITEREG((MRST_N << 16) , MC1);

    /* shut down all dma transfers and rps tasks */
//  WRITEREG(0x30ff0000, MC1);

    /* clear out any rps-signals pending */
//  WRITEREG(0xf8000000, MC2);

    printk(KERN_ERR "setup chip\n");

    /* enable audio interface */
#ifdef __BIG_ENDIAN
    acon1 |= A1_SWAP;
    acon1 |= A2_SWAP;
#endif
    /* WS0_CTRL, WS0_SYNC: input TSL1, I2S */
    /* WS1_CTRL, WS1_SYNC: output TSL2, I2S */
    acon1 |= 2 * WS1_CTRL;
    /* WS2_CTRL, WS2_SYNC: output TSL1, I2S */
    acon1 |= 1 * WS2_CTRL;
    /* It seems that WS4 must no be 3-state for digital output. 
       Why ? Don't know :-( 

       But it must be 3-state for input
       Why ? Still don't know
       Here WS4 is forced active LOW */    
    //acon1 |= 3 * WS4_CTRL; 

    
    /* WS3_CTRL, WS3_SYNC: output TSL2, I2S */
    acon1 |= 2 * WS3_CTRL;
    /* WS4_CTRL, WS4_SYNC: output TSL2, I2S */
    //acon1 |= 2 * WS4_CTRL;

    /* ou 4* comme  l'origine, ou encore autre chose ??? */
    acon1 |= 3 * AUDIO_MODE;
    WRITEREG(acon1, ACON1);

    WRITEREG(3 * (BurstA1_in) + 2 * (ThreshA1_in), PCI_BT_A);
    WRITEREG(3 * (BurstA1_out) + 2 * (ThreshA1_out), PCI_BT_A);
    WRITEREG(3 * (BurstA2_out) + 2 * (ThreshA2_out), PCI_BT_A);
    //WRITEREG(0, PCI_BT_A);


    /* enable audio port pins */
    WRITEREG((EAP << 16) | EAP, MC1);

    /* enable I2C */
    WRITEREG((EI2C << 16) | EI2C, MC1);
    /* enable interrupts */
    WRITEREG(A1_out | A2_out | A1_in | IIC_S | IIC_E, IER);

#if 1

    /* audio configuration */
    acon2 = A2_CLKSRC | BCLK1_OEN;
    WRITEREG(acon2, ACON2);

    //WRITEREG(0x50, GPIO_CTRL); // Lent puis rapide
#else
    //acon2 = BCLK1_OEN | A2_CLKSRC*9 | A1_CLKSRC*9;
    //acon2 = BCLK1_OEN | A2_CLKSRC*0x11 | A1_CLKSRC*0x11;

    acon2 = BCLK1_OEN | A1_CLKSRC*0 | A2_CLKSRC*5 ; // *x5 -> identique

    //WRITEREG(0x00000050, GPIO_CTRL); // rapide

    WRITEREG(0x40505040, GPIO_CTRL); // rapide

    //acon2 = A2_CLKSRC | BCLK1_OEN;
    WRITEREG(acon2, ACON2);


    //WRITEREG(0x50505050, GPIO_CTRL); // rapide
    //WRITEREG(0x40404040, GPIO_CTRL); // Lent puis rapide
    //WRITEREG(0x40405040, GPIO_CTRL); // Lent puis rapide
    //WRITEREG(0x50404040, GPIO_CTRL); // Lent puis rapide

#endif

    /* TSL setup */
    for (i = 0; i < 8; ++i)
    {
        WRITEREG(tsl1[i], TSL1 + (i * 4));
        WRITEREG(tsl2[i], TSL2 + (i * 4));
    }


/*
  iicsta = READREG(IICSTA);
  printk(KERN_ALERT "IICSTA: %X\n", iicsta);
  iicsta = 3 * IICCC;*/ /* PCI clock/320 */
/*  WRITEREG(ABORT, IICSTA);
    WRITEREG((UPLD_IIC << 16) | UPLD_IIC,  MC2);*/
    /* clear error bits (2x) */
/*  WRITEREG(iicsta,  IICSTA);
    WRITEREG((UPLD_IIC << 16) | UPLD_IIC,  MC2);
    WRITEREG(iicsta,  IICSTA);
    WRITEREG((UPLD_IIC << 16) | UPLD_IIC,   MC2);
    iicsta = READREG(IICSTA);
    printk(KERN_ALERT "IICSTA: %X\n", iicsta);
*/

    //dev->limit = 0;
}

void snd_aw2_saa7146_pcm_init_playback(snd_aw2_saa7146_t *chip, int stream_number, unsigned long dma_addr, 
                                        unsigned long period_size, unsigned long buffer_size)
{
    unsigned long dwPage, dwLimit;

    //aw2_set_sample_rate(chip, runtime->rate);
	
    /* @todo : on pense que ce n'est pas utile -> a confirmer peut tre par la suite
       snd_aw2_saa7146_set_channels(chip, runtime->channels); */

    // Configure DMA for substream
    // Configuration informations:
    //  - ALSA has allocated continuous memory pages. So we don't need to use MMU from saa7146.


    // No MMU -> nothing to do with PageA1, we only configure the limit of PageAx_out register
    // Disable MMU
    dwPage = (0L<<11);
    /* Configure Limit for DMA access 

    The limit register defines an address limit, which generates
    an interrupt if passed by the actual PCI address pointer.
    '0001' means an interrupt will be generated if the lower
    6 bits (64 bytes) of the PCI address are zero. '0010'
    defines a limit of 128 bytes, '0011' one of 256 bytes, and
    so on up to 1 Mbyte defined by '1111'. This interrupt range
    can be calculated as follows:
    Range = 2^(5 + Limit) bytes.
    */
    dwLimit = snd_aw2_saa7146_get_limit(period_size);
    dwPage |= (dwLimit << 4);

    //printk("dwLimit=0x%x\n",dwLimit);

    if (stream_number == 0)
    {
        WRITEREG(dwPage, PageA1_out);

        // Base address for DMA transfert.
        // This address has been reserved by ALSA.
        // This is a physical address
        WRITEREG(dma_addr, BaseA1_out);

        // Define upper limit for DMA access        
        WRITEREG(dma_addr + buffer_size, ProtA1_out);

    }
    else if (stream_number == 1)
    {
        WRITEREG(dwPage, PageA2_out);

        // Base address for DMA transfert.
        // This address has been reserved by ALSA.
        // This is a physical address
        WRITEREG(dma_addr, BaseA2_out);

        // Define upper limit for DMA access        
        WRITEREG(dma_addr + buffer_size, ProtA2_out);
    }
    else
    {
        printk(KERN_ALERT "snd_aw2_saa7146_pcm_init_playback: Substream number is not 0 or 1 -> not managed, ERROR!!!\n");
    }
}

void snd_aw2_saa7146_pcm_init_capture(snd_aw2_saa7146_t *chip, int stream_number, unsigned long dma_addr, 
                                        unsigned long period_size, unsigned long buffer_size)
{
    unsigned long dwPage, dwLimit;

    //aw2_set_sample_rate(chip, runtime->rate);
	
    /* @todo : on pense que ce n'est pas utile -> a confirmer peut tre par la suite
       snd_aw2_saa7146_set_channels(chip, runtime->channels); */

    // Configure DMA for substream
    // Configuration informations:
    //  - ALSA has allocated continuous memory pages. So we don't need to use MMU from saa7146.


    // No MMU -> nothing to do with PageA1, we only configure the limit of PageAx_in register
    // Disable MMU
    dwPage = (0L<<11);
    /* Configure Limit for DMA access 

    The limit register defines an address limit, which generates
    an interrupt if passed by the actual PCI address pointer.
    '0001' means an interrupt will be generated if the lower
    6 bits (64 bytes) of the PCI address are zero. '0010'
    defines a limit of 128 bytes, '0011' one of 256 bytes, and
    so on up to 1 Mbyte defined by '1111'. This interrupt range
    can be calculated as follows:
    Range = 2^(5 + Limit) bytes.
    */
    dwLimit = snd_aw2_saa7146_get_limit(period_size);
    dwPage |= (dwLimit << 4);

    if (stream_number == 0)
    {
        WRITEREG(dwPage, PageA1_in);

        // Base address for DMA transfert.
        // This address has been reserved by ALSA.
        // This is a physical address
        WRITEREG(dma_addr, BaseA1_in);

        // Define upper limit for DMA access        
        WRITEREG(dma_addr + buffer_size, ProtA1_in);

    }
    else
    {
        printk(KERN_ALERT "snd_aw2_saa7146_pcm_capture: Substream number is not 0 -> not managed, ERROR!!!\n");
    }
}

void snd_aw2_saa7146_define_it_playback_callback(unsigned int stream_number, snd_aw2_saa7146_it_cb pItCallback, void * pCallbackParam)
{
    if (stream_number < NB_STREAM_PLAYBACK)
    {
        arrSubstreamItPlaybackCb[stream_number].pItCallback = (snd_aw2_saa7146_it_cb)pItCallback;
        arrSubstreamItPlaybackCb[stream_number].pCallbackParam = (void*)pCallbackParam;
    }
}

void snd_aw2_saa7146_define_it_capture_callback(unsigned int stream_number, snd_aw2_saa7146_it_cb pItCallback, void * pCallbackParam)
{
    if (stream_number < NB_STREAM_CAPTURE)
    {
        arrSubstreamItCaptureCb[stream_number].pItCallback = (snd_aw2_saa7146_it_cb)pItCallback;
        arrSubstreamItCaptureCb[stream_number].pCallbackParam = (void*)pCallbackParam;
    }
}

void snd_aw2_saa7146_pcm_trigger_start_playback(snd_aw2_saa7146_t *chip, int stream_number)
{
    printk(KERN_ALERT "snd_aw2_saa7146_pcm_trigger: START - substream->number=%x\n",stream_number);
    // In aw8 driver, dma transfert is always active. It is
    // started and stopped in a larger "space"
    if (stream_number == 0)
    {
		
        WRITEREG((TR_E_A1_OUT << 16) | TR_E_A1_OUT, MC1);
    }
    else if (stream_number == 1)
    {
        WRITEREG((TR_E_A2_OUT << 16) | TR_E_A2_OUT,  MC1);
    }
    else 
    {
    }


}

void snd_aw2_saa7146_pcm_trigger_stop_playback(snd_aw2_saa7146_t *chip, int stream_number)
{
    printk(KERN_ALERT "snd_aw2_saa7146_pcm_trigger: STOP - substream->number=%x\n",stream_number);
    //WRITEREG(0, ACON1);
    if (stream_number == 0)
    {
        WRITEREG((TR_E_A1_OUT << 16) , MC1);
    }
    else if (stream_number == 1)
    {
        WRITEREG((TR_E_A2_OUT << 16) ,  MC1);
    } 
    else
    {
    }
}


void snd_aw2_saa7146_pcm_trigger_start_capture(snd_aw2_saa7146_t *chip, int stream_number)
{
    printk(KERN_ALERT "snd_aw2_saa7146_pcm_trigger: START - substream->number=%x\n",stream_number);
    // In aw8 driver, dma transfert is always active. It is
    // started and stopped in a larger "space"
    if (stream_number == 0)
    {
		
        WRITEREG((TR_E_A1_IN << 16) | TR_E_A1_IN, MC1);
    }
    else
    {
    }
}

void snd_aw2_saa7146_pcm_trigger_stop_capture(snd_aw2_saa7146_t *chip, int stream_number)
{
    printk(KERN_ALERT "snd_aw2_saa7146_pcm_trigger: STOP - substream->number=%x\n",stream_number);
    //WRITEREG(0, ACON1);
    if (stream_number == 0)
    {
        WRITEREG((TR_E_A1_IN << 16) , MC1);
    }
    else
    {
    }

}

irqreturn_t snd_aw2_saa7146_interrupt(int irq, void *dev_id,
                                      struct pt_regs *regs)
{
    unsigned int isr;
    unsigned int iicsta;
    snd_aw2_saa7146_t *chip = dev_id;
  
    isr = READREG( ISR);
    if (! isr) 
        return IRQ_NONE;


    WRITEREG(isr,  ISR);

    if (isr & (IIC_S | IIC_E))
    {
        iicsta = READREG(IICSTA);
        WRITEREG(0x100, IICSTA);
        wake_up(&chip->wait_iic);
    }
    else if (isr)
    {
    }

    if (isr & A1_out )
    {
        if (arrSubstreamItPlaybackCb[0].pItCallback != NULL)
        {
            arrSubstreamItPlaybackCb[0].pItCallback(arrSubstreamItPlaybackCb[0].pCallbackParam);
        }
        //snd_pcm_period_elapsed(chip->playback_substream[0]);
    }
    if (isr & A2_out)
    {
        if (arrSubstreamItPlaybackCb[1].pItCallback != NULL)
        {
            arrSubstreamItPlaybackCb[1].pItCallback(arrSubstreamItPlaybackCb[1].pCallbackParam);
        }

        //snd_pcm_period_elapsed(chip->playback_substream[1]);
    }
#if 1
    if (isr & A1_in) 
    {
        if (arrSubstreamItCaptureCb[0].pItCallback != NULL)
        {
            arrSubstreamItCaptureCb[0].pItCallback(arrSubstreamItCaptureCb[0].pCallbackParam);
        }
    }
#else
    if (isr & A1_in) {
        spin_lock(&dev->lock);
        dev->pos += (1 << (5 + dev->limit)) / 8;
        spin_unlock(&dev->lock);
		
        wake_up(&dev->wait_xfer);
    }
#endif
    return IRQ_HANDLED;
}

unsigned int snd_aw2_saa7146_get_hw_pointer_playback(snd_aw2_saa7146_t *chip, int stream_number, 
                                                     unsigned char * startAddr, unsigned int bufferSize)
{
    int pci_adp = 0 ;
    size_t ptr = 0;

    if (stream_number == 0)
    {
        pci_adp = READREG(PCI_ADP1);
        ptr = pci_adp - (int)startAddr;

        if (ptr == bufferSize)
        {
            ptr = 0;
        }
    }
    if (stream_number == 1)
    {
        pci_adp = READREG(PCI_ADP3);
        ptr = pci_adp - (int)startAddr;

        if (ptr == bufferSize)
        {
            ptr = 0;
        }
    }
    else
    {
    }
//    printk("pointer stream:%d ptr:0x%x\n",stream_number, ptr);
    return ptr;
}

unsigned int snd_aw2_saa7146_get_hw_pointer_capture(snd_aw2_saa7146_t *chip, int stream_number,
                                                    unsigned char * startAddr, unsigned int bufferSize)
{
    int pci_adp = 0 ;
    size_t ptr = 0;
    if (stream_number == 0)
    {
        pci_adp = READREG(PCI_ADP2);
        ptr = pci_adp - (int)startAddr;

        if (ptr == bufferSize)
        {
            ptr = 0;
        }
    }
    else
    {
    }
    printk("pointer stream capture:%d ptr:0x%x\n",stream_number, ptr);
    return ptr;	 
}


int snd_aw2_saa7146_set_sample_rate(snd_aw2_saa7146_t * chip, int srate)
{
    unsigned int S;
    unsigned char DB[4];
    unsigned int iicsta = 0;
    int err=0;
    printk(KERN_ALERT "aw2_set_sample_rate: rate: %x\n", srate);
    /* check srate range */
    if ((srate < 30000) || (srate > 50000))
        return -EINVAL;

    /* PLL output is 256*fs, reference 1kHz */
    S = srate * 256 / 1000;

    /* LSB is at S2 */
    S <<= 2;

    DB[0] = (S << 1) & 0xff;
    DB[1] = (S >> 7) & 0xff;
    DB[2] = ((S >> 15) & 0xff) | 0x18;
    DB[3] = 0;

    if(chip->iic)
        return -EBUSY;
    chip->iic = 1;
    iicsta = READREG(IICSTA);
    printk(KERN_ALERT "On va appeler I2C WRITE [%x] DB[%x-%x-%x-%x]\n",
           iicsta,DB[0],DB[1],DB[2],DB[3]);
    up(&chip->sem);
    err = snd_aw2_saa7146_i2c_write(chip, 0xc4, DB, sizeof(DB));
    down(&chip->sem);
    chip->iic = 0;
    return err;

}

static int snd_aw2_saa7146_i2c_write(snd_aw2_saa7146_t* chip, unsigned char address,
                                     const void* buf, size_t count) 
{
    unsigned int iicsta = 0;
    unsigned int iictfr = 0;
    int i, b;

    iicsta = (READREG(IICSTA) & 0xFF); // On ne garde que les 8 bits de droite, cad sans le paramtrage de la clock

    // Paramtrage de la clock:
#if 1
    iicsta = 1 * IICCC; /* PCI clock/3200 */ // On met 1 comme dans le driver Windows
#else
    iicsta = 3 * IICCC; /* PCI clock/320 */
#endif
    WRITEREG(iicsta,  IICSTA);
    WRITEREG((UPLD_IIC << 16) | UPLD_IIC,  MC2);
    WRITEREG(iicsta,  IICSTA);
    WRITEREG((UPLD_IIC << 16) | UPLD_IIC,  MC2);
    if ((iicsta & BUSY) || (iicsta & ERR))
    { /* error or busy */
        printk(KERN_ALERT "snd_aw2_saa7146_i2c_write: [%s %s] ERROR OR BUSY !! \n",__DATE__,__TIME__);
        WRITEREG(iicsta | ABORT, IICSTA);
        WRITEREG((UPLD_IIC << 16) | UPLD_IIC,  MC2);
        /* clear error bits (2x) */
        WRITEREG(iicsta,  IICSTA);
        WRITEREG((UPLD_IIC << 16) | UPLD_IIC,  MC2);
        WRITEREG(iicsta,  IICSTA);
        WRITEREG((UPLD_IIC << 16) | UPLD_IIC,  MC2);
    }
// DEBUG
    iicsta = READREG(IICSTA);
    printk(KERN_ALERT "snd_aw2_saa7146_i2c_write: [%s %s] IICSTA after clear iicsta:[%x] !! \n",__DATE__,__TIME__,iicsta);
//////////

    iictfr = address * BYTE2;
    iictfr |= START * ATRR2;

    for (i = 0; i <= count; ++i)
    {
        printk(KERN_ALERT "snd_aw2_saa7146_i2c_write: [%s %s] loop counter:[%d]\n",__DATE__,__TIME__,i);
        b = i % 3;
        if ((b == 0) || (i == count))
        {
            wait_queue_t wait;
            init_waitqueue_entry(&wait, current);

            add_wait_queue(&chip->wait_iic, &wait);
            set_current_state(TASK_INTERRUPTIBLE);

            WRITEREG(iictfr,  IICTFR);
            WRITEREG((UPLD_IIC << 16) | UPLD_IIC,  MC2);
			
            printk(KERN_ALERT "snd_aw2_saa7146_i2c_write: [%s %s] before scheduling in lopp iictfr[%x] sizeof(iictfr)[%d]\n",__DATE__,__TIME__,iictfr,sizeof(iictfr));
            schedule();
            printk(KERN_ALERT "snd_aw2_saa7146_i2c_write: [%s %s] afer scheduling\n",__DATE__,__TIME__);
            set_current_state(TASK_RUNNING);
            remove_wait_queue(&chip->wait_iic, &wait);

            if (signal_pending(current))
            {
                printk(KERN_ALERT "snd_aw2_saa7146_i2c_write: [%s %s] *** I2C return SIGNAL_PENDING\n",__DATE__,__TIME__);
                return -ERESTARTSYS;
            }

            iicsta = READREG( IICSTA);
            if (iicsta & ERR) {
                printk(KERN_ALERT "snd_aw2_saa7146_i2c_write: [%s %s] *** I2C return PB IEO IICSTA[%x] \n",__DATE__,__TIME__,iicsta);
                return -EIO;
            }
            if (i == count) break;
            iictfr = 0;
        }
        switch (b)
        {
            case 0:
                iictfr |= BYTE2 * ((char*)buf)[i];
                iictfr |= ATRR2 * ((count - i == 1)?STOP:CONT);
                break;
            case 1:
                iictfr |= BYTE1 * ((char*)buf)[i];
                iictfr |= ATRR1 * ((count - i == 1)?STOP:CONT);
                break;
            case 2:
                iictfr |= BYTE0 * ((char*)buf)[i];
                iictfr |= ATRR0 * ((count - i == 1)?STOP:CONT);
                break;
        }
    }
    printk(KERN_ALERT "snd_aw2_saa7146_i2c_write: [%s %s] I2C termined (with success we hope...)\n",__DATE__,__TIME__);
    return 0;
}


static int snd_aw2_saa7146_get_limit(int size)
{
    int limitsize = 32;
    int limit = 0;
    while( limitsize < size ) {
	limitsize *= 2;
	limit++;
    }
    return limit;
}
